@@ -1,5 +1,5 @@
 /* xgettext Lua backend.
-   Copyright (C) 2012-2013, 2016, 2018-2019 Free Software Foundation, Inc.
+   Copyright (C) 2012-2024 Free Software Foundation, Inc.
 
    This file was written by Ľubomír Remák <lubomirr@lubomirr.eu>, 2012.
 
@@ -28,6 +28,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 
+#include <error.h>
+#include "attribute.h"
 #include "message.h"
 #include "rc-str-list.h"
 #include "xgettext.h"
@@ -37,8 +39,9 @@
 #include "xg-arglist-callshape.h"
 #include "xg-arglist-parser.h"
 #include "xg-message.h"
-#include "error.h"
+#include "if-error.h"
 #include "xalloc.h"
+#include "string-buffer.h"
 #include "gettext.h"
 #include "po-charset.h"
 
@@ -138,7 +141,7 @@ static FILE *fp;
 static unsigned char phase1_pushback[2];
 static int phase1_pushback_length;
 
-static int first_character = 1;
+static bool first_character;
 
 static int
 phase1_getc ()
@@ -153,7 +156,7 @@ phase1_getc ()
 
       if (first_character)
         {
-          first_character = 0;
+          first_character = false;
 
           /* Ignore shebang line.  No pushback required in this case.  */
           if (c == '#')
@@ -308,42 +311,41 @@ phase2_getc ()
                         break;
 
                       /* Ignore leading spaces and tabs.  */
-                      if (buflen == 0 && (c == ' ' || c == '\t'))
-                        continue;
-
-                      comment_add (c);
-
-                      switch (c)
+                      if (!(buflen == 0 && (c == ' ' || c == '\t')))
                         {
-                        case ']':
-                          if (!right_bracket)
+                          comment_add (c);
+
+                          switch (c)
                             {
-                              right_bracket = true;
-                              esigns2 = 0;
-                            }
-                          else
-                            {
-                              if (esigns2 == esigns)
+                            case ']':
+                              if (!right_bracket)
                                 {
-                                  comment_line_end (2 + esigns);
-                                  end = true;
+                                  right_bracket = true;
+                                  esigns2 = 0;
                                 }
+                              else
+                                {
+                                  if (esigns2 == esigns)
+                                    {
+                                      comment_line_end (2 + esigns);
+                                      end = true;
+                                    }
+                                }
+                              break;
+
+                            case '=':
+                              if (right_bracket)
+                                esigns2++;
+                              break;
+
+                            case '\n':
+                              comment_line_end (1);
+                              comment_start ();
+                              lineno = line_number;
+                              FALLTHROUGH;
+                            default:
+                              right_bracket = false;
                             }
-                          break;
-
-                        case '=':
-                          if (right_bracket)
-                            esigns2++;
-                          break;
-
-                        case '\n':
-                          comment_line_end (1);
-                          comment_start ();
-                          lineno = line_number;
-                          /* Intentionally not breaking.  */
-
-                        default:
-                          right_bracket = false;
                         }
                     }
                   last_comment_line = lineno;
@@ -427,41 +429,6 @@ free_token (token_ty *tp)
     drop_reference (tp->comment);
 }
 
-/* Our current string.  */
-static int string_buf_length;
-static int string_buf_alloc;
-static char *string_buf;
-
-static void
-string_start ()
-{
-  string_buf_length = 0;
-}
-
-static void
-string_add (int c)
-{
-  if (string_buf_length >= string_buf_alloc)
-    {
-      string_buf_alloc = 2 * string_buf_alloc + 10;
-      string_buf = xrealloc (string_buf, string_buf_alloc);
-    }
-
-  string_buf[string_buf_length++] = c;
-}
-
-static void
-string_end ()
-{
-  if (string_buf_length >= string_buf_alloc)
-    {
-      string_buf_alloc = string_buf_alloc + 1;
-      string_buf = xrealloc (string_buf, string_buf_alloc);
-    }
-
-  string_buf[string_buf_length] = '\0';
-}
-
 
 /* We need 3 pushback tokens for string optimization.  */
 static int phase3_pushback_length;
@@ -508,7 +475,7 @@ phase3_get (token_ty *tp)
         case '\n':
           if (last_non_comment_line > last_comment_line)
             savable_comment_reset ();
-          /* Intentionally not breaking.  */
+          FALLTHROUGH;
         case ' ':
         case '\t':
         case '\f':
@@ -592,124 +559,126 @@ phase3_get (token_ty *tp)
 
         case '"':
         case '\'':
-          c_start = c;
-          string_start ();
+          {
+            c_start = c;
+            struct string_buffer buffer;
+            sb_init (&buffer);
 
-          for (;;)
-            {
-              /* We need unprocessed characters from phase 1.  */
-              c = phase1_getc ();
+            for (;;)
+              {
+                /* We need unprocessed characters from phase 1.  */
+                c = phase1_getc ();
 
-              if (c == EOF || c == c_start || c == '\n')
-                {
-                  /* End of string.  */
-                  string_end ();
-                  tp->string = xstrdup (string_buf);
-                  tp->comment = add_reference (savable_comment);
-                  tp->type = token_type_string;
-                  return;
-                }
+                if (c == EOF || c == c_start || c == '\n')
+                  {
+                    /* End of string.  */
+                    tp->string = sb_xdupfree_c (&buffer);
+                    tp->comment = add_reference (savable_comment);
+                    tp->type = token_type_string;
+                    return;
+                  }
 
-              /* We got '\', this is probably an escape sequence.  */
-              if (c == '\\')
-                {
-                  c = phase1_getc ();
-                  switch (c)
-                    {
-                    case 'a':
-                      string_add ('\a');
-                      break;
-                    case 'b':
-                      string_add ('\b');
-                      break;
-                    case 'f':
-                      string_add ('\f');
-                      break;
-                    case 'n':
-                      string_add ('\n');
-                      break;
-                    case 'r':
-                      string_add ('\r');
-                      break;
-                    case 't':
-                      string_add ('\t');
-                      break;
-                    case 'v':
-                      string_add ('\v');
-                      break;
-                    case 'x':
+                /* We got '\', this is probably an escape sequence.  */
+                if (c == '\\')
+                  {
+                    c = phase1_getc ();
+                    switch (c)
                       {
-                        int num = 0;
-                        int i = 0;
-
-                        for (i = 0; i < 2; i++)
-                          {
-                            c = phase1_getc ();
-                            if (c >= '0' && c <= '9')
-                              num += c - '0';
-                            else if (c >= 'a' && c <= 'f')
-                              num += c - 'a' + 10;
-                            else if (c >= 'A' && c <= 'F')
-                              num += c - 'A' + 10;
-                            else
-                              {
-                                phase1_ungetc (c);
-                                break;
-                              }
-
-                            if (i == 0)
-                              num *= 16;
-                          }
-
-                        if (i == 2)
-                          string_add (num);
-                      }
-
-                      break;
-                    case 'z':
-                      /* Ignore the following whitespace.  */
-                      do
-                        {
-                          c = phase1_getc ();
-                        }
-                      while (c == ' ' || c == '\n' || c == '\t' || c == '\r'
-                             || c == '\f' || c == '\v');
-
-                      phase1_ungetc (c);
-
-                      break;
-                    default:
-                      /* Check if it's a '\ddd' sequence.  */
-                      if (c >= '0' && c <= '9')
+                      case 'a':
+                        sb_xappend1 (&buffer, '\a');
+                        break;
+                      case 'b':
+                        sb_xappend1 (&buffer, '\b');
+                        break;
+                      case 'f':
+                        sb_xappend1 (&buffer, '\f');
+                        break;
+                      case 'n':
+                        sb_xappend1 (&buffer, '\n');
+                        break;
+                      case 'r':
+                        sb_xappend1 (&buffer, '\r');
+                        break;
+                      case 't':
+                        sb_xappend1 (&buffer, '\t');
+                        break;
+                      case 'v':
+                        sb_xappend1 (&buffer, '\v');
+                        break;
+                      case 'x':
                         {
                           int num = 0;
                           int i = 0;
 
-                          while (c >= '0' && c <= '9' && i < 3)
+                          for (i = 0; i < 2; i++)
                             {
-                              num *= 10;
-                              num += (c - '0');
                               c = phase1_getc ();
-                              i++;
+                              if (c >= '0' && c <= '9')
+                                num += c - '0';
+                              else if (c >= 'a' && c <= 'f')
+                                num += c - 'a' + 10;
+                              else if (c >= 'A' && c <= 'F')
+                                num += c - 'A' + 10;
+                              else
+                                {
+                                  phase1_ungetc (c);
+                                  break;
+                                }
+
+                              if (i == 0)
+                                num *= 16;
                             }
 
-                          /* The last read character is either a
-                             non-number or another number after our
-                             '\ddd' sequence.  We need to ungetc it.  */
-                          phase1_ungetc (c);
-
-                          /* The sequence number is too big, this
-                             causes a lexical error.  Ignore it.  */
-                          if (num < 256)
-                            string_add (num);
+                          if (i == 2)
+                            sb_xappend1 (&buffer, num);
                         }
-                      else
-                        string_add (c);
-                    }
-                }
-              else
-                string_add (c);
-            }
+
+                        break;
+                      case 'z':
+                        /* Ignore the following whitespace.  */
+                        do
+                          {
+                            c = phase1_getc ();
+                          }
+                        while (c == ' ' || c == '\n' || c == '\t' || c == '\r'
+                               || c == '\f' || c == '\v');
+
+                        phase1_ungetc (c);
+
+                        break;
+                      default:
+                        /* Check if it's a '\ddd' sequence.  */
+                        if (c >= '0' && c <= '9')
+                          {
+                            int num = 0;
+                            int i = 0;
+
+                            while (c >= '0' && c <= '9' && i < 3)
+                              {
+                                num *= 10;
+                                num += (c - '0');
+                                c = phase1_getc ();
+                                i++;
+                              }
+
+                            /* The last read character is either a
+                               non-number or another number after our
+                               '\ddd' sequence.  We need to ungetc it.  */
+                            phase1_ungetc (c);
+
+                            /* The sequence number is too big, this
+                               causes a lexical error.  Ignore it.  */
+                            if (num < 256)
+                              sb_xappend1 (&buffer, num);
+                          }
+                        else
+                          sb_xappend1 (&buffer, c);
+                      }
+                  }
+                else
+                  sb_xappend1 (&buffer, c);
+              }
+          }
           break;
 
         case '[':
@@ -741,68 +710,69 @@ phase3_get (token_ty *tp)
             }
 
           /* Found an opening long bracket.  */
-          string_start ();
+          {
+            struct string_buffer buffer;
+            sb_init (&buffer);
 
-          /* See if it is immediately followed by a newline.  */
-          c = phase1_getc ();
-          if (c != '\n')
-            phase1_ungetc (c);
+            /* See if it is immediately followed by a newline.  */
+            c = phase1_getc ();
+            if (c != '\n')
+              phase1_ungetc (c);
 
-          for (;;)
-            {
-              c = phase1_getc ();
+            for (;;)
+              {
+                c = phase1_getc ();
 
-              if (c == EOF)
-                {
-                  string_end ();
-                  tp->string = xstrdup (string_buf);
-                  tp->comment = add_reference (savable_comment);
-                  tp->type = token_type_string;
-                  return;
-                }
-              if (c == ']')
-                {
-                  c = phase1_getc ();
+                if (c == EOF)
+                  {
+                    tp->string = sb_xdupfree_c (&buffer);
+                    tp->comment = add_reference (savable_comment);
+                    tp->type = token_type_string;
+                    return;
+                  }
+                if (c == ']')
+                  {
+                    c = phase1_getc ();
 
-                  /* Count the number of equal signs.  */
-                  int esigns2 = 0;
-                  while (c == '=')
-                    {
-                      esigns2++;
-                      c = phase1_getc ();
-                    }
+                    /* Count the number of equal signs.  */
+                    int esigns2 = 0;
+                    while (c == '=')
+                      {
+                        esigns2++;
+                        c = phase1_getc ();
+                      }
 
-                  if (c == ']' && esigns == esigns2)
-                    {
-                      /* We got ']==...==]', where the number of equal
-                         signs matches the number of equal signs in
-                         the opening bracket.  */
-                      string_end ();
-                      tp->string = xstrdup (string_buf);
-                      tp->comment = add_reference (savable_comment);
-                      tp->type = token_type_string;
-                      return;
-                    }
-                  else
-                    {
-                      /* Otherwise we got either ']==' garbage or
-                         ']==...==]' with a different number of equal
-                         signs.
+                    if (c == ']' && esigns == esigns2)
+                      {
+                        /* We got ']==...==]', where the number of equal
+                           signs matches the number of equal signs in
+                           the opening bracket.  */
+                        tp->string = sb_xdupfree_c (&buffer);
+                        tp->comment = add_reference (savable_comment);
+                        tp->type = token_type_string;
+                        return;
+                      }
+                    else
+                      {
+                        /* Otherwise we got either ']==' garbage or
+                           ']==...==]' with a different number of equal
+                           signs.
 
-                         Add ']' and equal signs to the string, and
-                         ungetc the current character, because the
-                         second ']' might be a part of another closing
-                         long bracket, e.g. '==]===]'.  */
-                      phase1_ungetc (c);
+                           Add ']' and equal signs to the string, and
+                           ungetc the current character, because the
+                           second ']' might be a part of another closing
+                           long bracket, e.g. '==]===]'.  */
+                        phase1_ungetc (c);
 
-                      string_add (']');
-                      while (esigns2--)
-                        string_add ('=');
-                    }
-                }
-              else
-                string_add (c);
-            }
+                        sb_xappend1 (&buffer, ']');
+                        while (esigns2--)
+                          sb_xappend1 (&buffer, '=');
+                      }
+                  }
+                else
+                  sb_xappend1 (&buffer, c);
+              }
+          }
           break;
 
         case ']':
@@ -838,25 +808,36 @@ phase3_get (token_ty *tp)
           else if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
                    || c == '_')
             {
-              string_start ();
+              struct string_buffer buffer;
+              sb_init (&buffer);
+
               while ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
                      || c == '_' || (c >= '0' && c <= '9'))
                 {
-                  string_add (c);
+                  sb_xappend1 (&buffer, c);
                   c = phase1_getc ();
                 }
-              string_end ();
               phase1_ungetc (c);
 
-              if (strcmp (string_buf, "not") == 0)
-                tp->type = token_type_operator1;
-              else if (strcmp (string_buf, "and") == 0)
-                tp->type = token_type_operator2;
-              else if (strcmp (string_buf, "or") == 0)
-                tp->type = token_type_operator2;
+              const char *contents = sb_xcontents_c (&buffer);
+              if (strcmp (contents, "not") == 0)
+                {
+                  sb_free (&buffer);
+                  tp->type = token_type_operator1;
+                }
+              else if (strcmp (contents, "and") == 0)
+                {
+                  sb_free (&buffer);
+                  tp->type = token_type_operator2;
+                }
+              else if (strcmp (contents, "or") == 0)
+                {
+                  sb_free (&buffer);
+                  tp->type = token_type_operator2;
+                }
               else
                 {
-                  tp->string = xstrdup (string_buf);
+                  tp->string = sb_xdupfree_c (&buffer);
                   tp->type = token_type_symbol;
                 }
               return;
@@ -1010,6 +991,14 @@ x_lua_lex (token_ty *tok)
 static flag_context_list_table_ty *flag_context_list_table;
 
 
+/* Maximum supported nesting depth.  */
+#define MAX_NESTING_DEPTH 1000
+
+/* Current nesting depths.  */
+static int paren_nesting_depth;
+static int bracket_nesting_depth;
+
+
 /* The file is broken into tokens.  Scan the token stream, looking for
    a keyword, followed by a left paren, followed by a string.  When we
    see this sequence, we have something to remember.  We assume we are
@@ -1032,7 +1021,7 @@ static flag_context_list_table_ty *flag_context_list_table;
    Return true upon eof, false upon closing parenthesis or bracket.  */
 static bool
 extract_balanced (message_list_ty *mlp, token_type_ty delim,
-                  flag_context_ty outer_context,
+                  flag_region_ty *outer_region,
                   flag_context_list_iterator_ty context_iter,
                   struct arglist_parser *argparser)
 {
@@ -1045,9 +1034,9 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
   /* Context iterator that will be used if the next token is a '('.  */
   flag_context_list_iterator_ty next_context_iter =
     passthrough_context_list_iterator;
-  /* Current context.  */
-  flag_context_ty inner_context =
-    inherited_context (outer_context,
+  /* Current region.  */
+  flag_region_ty *inner_region =
+    inheriting_region (outer_region,
                        flag_context_list_iterator_advance (&context_iter));
 
   /* Start state is 0.  */
@@ -1084,14 +1073,20 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
           continue;
 
         case token_type_lparen:
+          if (++paren_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open parentheses"));
           if (extract_balanced (mlp, token_type_rparen,
-                                inner_context, next_context_iter,
+                                inner_region, next_context_iter,
                                 arglist_parser_alloc (mlp,
                                                       state ? next_shapes : NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          paren_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           break;
@@ -1100,6 +1095,7 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
           if (delim == token_type_rparen || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
 
@@ -1108,13 +1104,20 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
           continue;
 
         case token_type_lbracket:
+          if (++bracket_nesting_depth > MAX_NESTING_DEPTH)
+            if_error (IF_SEVERITY_FATAL_ERROR,
+                      logical_file_name, line_number, (size_t)(-1), false,
+                      _("too many open brackets"));
           if (extract_balanced (mlp, token_type_rbracket,
-                                null_context, null_context_list_iterator,
+                                null_context_region (),
+                                null_context_list_iterator,
                                 arglist_parser_alloc (mlp, NULL)))
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return true;
             }
+          bracket_nesting_depth--;
           next_context_iter = null_context_list_iterator;
           state = 0;
           break;
@@ -1123,6 +1126,7 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
           if (delim == token_type_rbracket || delim == token_type_eof)
             {
               arglist_parser_done (argparser, arg);
+              unref_region (inner_region);
               return false;
             }
 
@@ -1132,8 +1136,9 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
 
         case token_type_comma:
           arg++;
-          inner_context =
-            inherited_context (outer_context,
+          unref_region (inner_region);
+          inner_region =
+            inheriting_region (outer_region,
                                flag_context_list_iterator_advance (
                                  &context_iter));
           next_context_iter = passthrough_context_list_iterator;
@@ -1142,6 +1147,7 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
 
         case token_type_eof:
           arglist_parser_done (argparser, arg);
+          unref_region (inner_region);
           return true;
 
         case token_type_string:
@@ -1151,8 +1157,9 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
             pos.line_number = token.line_number;
 
             if (extract_all)
-              remember_a_message (mlp, NULL, token.string, false, inner_context,
-                                  &pos, NULL, token.comment, false);
+              remember_a_message (mlp, NULL, token.string, false, false,
+                                  inner_region, &pos,
+                                  NULL, token.comment, false);
             else
               {
                 mixed_string_ty *ms =
@@ -1166,14 +1173,14 @@ extract_balanced (message_list_ty *mlp, token_type_ty delim,
                     tmp_argparser = arglist_parser_alloc (mlp, next_shapes);
 
                     arglist_parser_remember (tmp_argparser, 1, ms,
-                                             inner_context,
+                                             inner_region,
                                              pos.file_name, pos.line_number,
                                              token.comment, false);
                     arglist_parser_done (tmp_argparser, 1);
                   }
                 else
                   arglist_parser_remember (argparser, arg, ms,
-                                           inner_context,
+                                           inner_region,
                                            pos.file_name, pos.line_number,
                                            token.comment, false);
               }
@@ -1212,17 +1219,27 @@ extract_lua (FILE *f,
   logical_file_name = xstrdup (logical_filename);
   line_number = 1;
 
+  phase1_pushback_length = 0;
+  first_character = true;
+
   last_comment_line = -1;
   last_non_comment_line = -1;
 
+  phase3_pushback_length = 0;
+
+  phase4_last = token_type_eof;
+  phase4_pushback_length = 0;
+
   flag_context_list_table = flag_table;
+  paren_nesting_depth = 0;
+  bracket_nesting_depth = 0;
 
   init_keywords ();
 
   /* Eat tokens until eof is seen.  When extract_parenthesized returns
      due to an unbalanced closing parenthesis, just restart it.  */
   while (!extract_balanced (mlp, token_type_eof,
-                            null_context, null_context_list_iterator,
+                            null_context_region (), null_context_list_iterator,
                             arglist_parser_alloc (mlp, NULL)))
     ;
 
